from decimal import Decimal
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from jacc.models import Invoice, INVOICE_CREDIT_NOTE, INVOICE_DEFAULT


def validate_invoice_settlement_amount(inv: Invoice, amt: Decimal, field_name: str = "amount"):
    """
    Validates that settlement amount is suitable to be used to settle specific invoice.
    Args:
        inv: Invoice
        amt: Decimal, amount
        field_name: Settlement account entry amount field, default "amount"

    Returns:
        None
    """
    if amt is None:
        raise ValidationError({field_name: _("Field value missing")})
    if amt == Decimal("0.00"):
        raise ValidationError({field_name: _("Amount cannot be zero")})
    if inv.type == INVOICE_DEFAULT and amt < Decimal("0.00"):
        raise ValidationError({field_name: _("Debit note settlement amount cannot be negative")})
    if inv.type == INVOICE_CREDIT_NOTE and amt > Decimal("0.00"):
        raise ValidationError({field_name: _("Credit note settlement amount cannot be positive")})
    unpaid_amt = inv.unpaid_amount or Decimal("0.00")
    if abs(amt) > abs(unpaid_amt):
        raise ValidationError({field_name: _("Settlement amount exceeds invoice unpaid amount")})



# Generated by Django 4.1.7 on 2023-05-03 19:28

from django.db import migrations
import jutil.modelfields


class Migration(migrations.Migration):
    dependencies = [
        ("jacc", "0028_alter_invoice_amount_alter_invoice_overpaid_amount_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="currency",
            field=jutil.modelfields.SafeCharField(blank=True, default="EUR", max_length=3, verbose_name="currency"),
        ),
    ]



from typing import List


def align_lines(lines: list, column_separator: str = "|") -> List[str]:
    """Pads lines so that all rows in single column match. Columns separated by '|' in every line.

    Args:
        lines: list of lines
        column_separator: column separator. default is '|'

    Returns:
        list of lines
    """
    rows = []
    col_len: List[int] = []
    for line in lines:
        line = str(line)
        cols = []
        for col_index, col in enumerate(line.split(column_separator)):
            col = str(col).strip()
            cols.append(col)
            if col_index >= len(col_len):
                col_len.append(0)
            col_len[col_index] = max(col_len[col_index], len(col))
        rows.append(cols)

    lines_out: List[str] = []
    for row in rows:
        cols_out = []
        for col_index, col in enumerate(row):
            if col_index == 0:
                col = col.ljust(col_len[col_index])
            else:
                col = col.rjust(col_len[col_index])
            cols_out.append(col)
        lines_out.append(" ".join(cols_out))
    return lines_out



# Generated by Django 4.0.4 on 2022-05-05 22:22

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0027_alter_invoice_paid_amount"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="amount",
            field=models.DecimalField(
                blank=True, decimal_places=2, default=Decimal("0.00"), help_text="invoice.amount.help_text", max_digits=10, verbose_name="amount"
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="overpaid_amount",
            field=models.DecimalField(
                blank=True,
                db_index=True,
                decimal_places=2,
                default=Decimal("0.00"),
                editable=False,
                help_text="invoice.overpaid_amount.help_text",
                max_digits=10,
                null=True,
                verbose_name="overpaid amount",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="paid_amount",
            field=models.DecimalField(
                blank=True,
                db_index=True,
                decimal_places=2,
                default=Decimal("0.00"),
                help_text="invoice.paid_amount.help_text",
                max_digits=10,
                null=True,
                verbose_name="paid amount",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="unpaid_amount",
            field=models.DecimalField(
                blank=True,
                db_index=True,
                decimal_places=2,
                default=None,
                editable=False,
                help_text="invoice.unpaid_amount.help_text",
                max_digits=10,
                null=True,
                verbose_name="unpaid amount",
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-06-08 14:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("jacc", "0029_alter_account_currency"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="accountentry",
            index=models.Index(fields=["account", "created"], name="jacc_accoun_account_f79b79_idx"),
        ),
    ]



# Generated by Django 3.2 on 2021-04-24 00:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0023_accountentrynote"),
    ]

    operations = [
        migrations.AddField(
            model_name="account",
            name="notes",
            field=models.TextField(blank=True, default="", verbose_name="notes"),
        ),
    ]



# Generated by Django 2.0 on 2018-01-21 17:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0007_invoice_number"),
    ]

    operations = [
        migrations.RenameField(
            model_name="accounttype",
            old_name="is_debit",
            new_name="is_asset",
        ),
        migrations.AlterField(
            model_name="accounttype",
            name="is_asset",
            field=models.BooleanField(verbose_name="is asset"),
        ),
    ]



from decimal import Decimal
from django.core.management.base import CommandParser
from jacc.models import AccountEntry, Invoice
from jutil.command import SafeCommand


class Command(SafeCommand):
    help = "Invoice balance"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("invoice", type=int)
        parser.add_argument("--tx", action="store_true")

    def do(self, *args, **options):
        inv = Invoice.objects.get(id=options["invoice"])
        assert isinstance(inv, Invoice)
        inv.update_cached_fields()

        print("Invoice id={} balances:".format(inv.id))
        for item, bal in inv.get_item_balances(inv.receivables_account):
            assert isinstance(item, AccountEntry)
            assert isinstance(bal, Decimal)
            print("  {} balance {}".format(item, bal))

        if options["tx"]:
            bal = Decimal("0.00")
            for tx in inv.receivables.order_by("timestamp", "id"):
                assert isinstance(tx, AccountEntry)
                bal += tx.amount
                print(
                    "  [{}] {} {} {}{} {}".format(
                        tx.id,
                        tx.timestamp.date().isoformat(),
                        tx.type,
                        "+" if tx.amount >= Decimal("0.00") else "",
                        tx.amount,
                        bal,
                    )
                )



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-10-30 19:58
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0004_auto_20171030_1935"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="accountentrysourcefile",
            options={"verbose_name": "account entry source file", "verbose_name_plural": "account entry source files"},
        ),
        migrations.AlterField(
            model_name="accountentry",
            name="source_file",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Source for account entry, e.g. received payment (if any)",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="jacc.AccountEntrySourceFile",
                verbose_name="account entry source file",
            ),
        ),
    ]



# Generated by Django 2.1.2 on 2018-10-18 15:36

from django.db import migrations, models
from django.db.models import F


def migr_code_to_identifier_0019_entrytype_identifier(apps, schema):
    EntryType = apps.get_model("jacc", "EntryType")
    EntryType.objects.all().update(identifier=F("code"))


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0018_auto_20181008_2322"),
    ]

    operations = [
        migrations.AddField(
            model_name="entrytype",
            name="identifier",
            field=models.CharField(blank=True, db_index=True, default="", max_length=40, verbose_name="identifier"),
        ),
        migrations.RunPython(migr_code_to_identifier_0019_entrytype_identifier),
    ]



from jacc.models import Invoice
from django.core.management.base import CommandParser
from jutil.command import SafeCommand


class Command(SafeCommand):
    help = "Updates cached values of invoices"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("--invoice", type=int)
        parser.add_argument("--force", action="store_true")
        parser.add_argument("--verbose", action="store_true")

    def do(self, *args, **options):
        invoices = Invoice.objects.all()
        if options["invoice"]:
            invoices = invoices.filter(id=options["invoice"])
        if not options["force"]:
            invoices = invoices.filter(close_date=None)

        count = 0
        for invoice in invoices.order_by("id"):
            assert isinstance(invoice, Invoice)
            if options["verbose"]:
                print("Updating", invoice)
            invoice.update_cached_fields()
            count += 1

        print("Updated", count, "invoices")



# Generated by Django 2.0.4 on 2018-05-05 15:12

from django.db import migrations, models


def make_entry_type_codes_unique(apps, schema_editor):
    EntryType = apps.get_model("jacc", "EntryType")
    for e in EntryType.objects.all().distinct("code"):
        dups = list(EntryType.objects.all().filter(code=e.code))
        if len(dups) > 1:
            for ix, e2 in enumerate(dups):
                if ix > 0:
                    e2.code += "({})".format(ix + 1)
                    e2.save(update_fields=["code"])


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0014_auto_20180430_1506"),
    ]

    operations = [
        migrations.RunPython(
            make_entry_type_codes_unique,
        ),
        migrations.AlterField(
            model_name="entrytype",
            name="code",
            field=models.CharField(db_index=True, max_length=32, unique=True, verbose_name="code"),
        ),
        migrations.AlterUniqueTogether(
            name="entrytype",
            unique_together=set(),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-10-12 23:29
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import jacc.models
import jutil.cache


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Account",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "currency",
                    models.CharField(
                        blank=True,
                        choices=[("EUR", "EUR"), ("USD", "USD")],
                        default="EUR",
                        max_length=3,
                        verbose_name="currency",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
            ],
            options={
                "verbose_name_plural": "accounts",
                "verbose_name": "account",
            },
        ),
        migrations.CreateModel(
            name="AccountEntry",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
                (
                    "timestamp",
                    models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now, verbose_name="timestamp"),
                ),
                ("description", models.CharField(blank=True, default="", max_length=256, verbose_name="description")),
                ("amount", models.DecimalField(decimal_places=2, max_digits=10, verbose_name="amount")),
                (
                    "account",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="accountentry_set",
                        to="jacc.Account",
                        verbose_name="account",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="child_set",
                        to="jacc.AccountEntry",
                        verbose_name="account.entry.parent",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "account entries",
                "verbose_name": "account entry",
            },
        ),
        migrations.CreateModel(
            name="AccountEntrySource",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(blank=True, db_index=True, default="", max_length=255, verbose_name="name")),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
            ],
            options={
                "verbose_name_plural": "account entry sources",
                "verbose_name": "account entry source",
            },
        ),
        migrations.CreateModel(
            name="AccountType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("code", models.CharField(db_index=True, max_length=32, unique=True, verbose_name="code")),
                ("name", models.CharField(db_index=True, max_length=64, unique=True, verbose_name="name")),
                ("is_debit", models.BooleanField(verbose_name="is debit")),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
            ],
            options={
                "verbose_name_plural": "account types",
                "verbose_name": "account type",
            },
        ),
        migrations.CreateModel(
            name="Contract",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
                ("name", models.CharField(blank=True, db_index=True, default="", max_length=128, verbose_name="name")),
            ],
            options={
                "verbose_name_plural": "contracts",
                "verbose_name": "contract",
            },
        ),
        migrations.CreateModel(
            name="EntryType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("code", models.CharField(db_index=True, max_length=32, unique=True, verbose_name="code")),
                ("name", models.CharField(db_index=True, max_length=64, unique=True, verbose_name="name")),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
                (
                    "payback_priority",
                    models.SmallIntegerField(blank=True, db_index=True, default=0, verbose_name="payback priority"),
                ),
                ("is_settlement", models.BooleanField(db_index=True, default=False, verbose_name="is settlement")),
                ("is_payment", models.BooleanField(db_index=True, default=False, verbose_name="is payment")),
            ],
            options={
                "verbose_name_plural": "entry types",
                "verbose_name": "entry type",
            },
        ),
        migrations.CreateModel(
            name="Invoice",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "type",
                    models.CharField(
                        blank=True,
                        choices=[("I1", "Invoice"), ("I2", "Credit Note")],
                        db_index=True,
                        default="I1",
                        max_length=2,
                        verbose_name="type",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, db_index=True, verbose_name="last modified")),
                ("sent", models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name="sent")),
                (
                    "due_date",
                    models.DateTimeField(db_index=True, default=jacc.models.get_default_due_date, verbose_name="due date"),
                ),
                ("notes", models.TextField(blank=True, default="", verbose_name="notes")),
                (
                    "filename",
                    models.CharField(blank=True, db_index=True, default="", max_length=255, verbose_name="filename"),
                ),
                (
                    "amount",
                    models.DecimalField(blank=True, decimal_places=2, default=0, max_digits=10, verbose_name="amount"),
                ),
                (
                    "paid_amount",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        default=None,
                        editable=False,
                        max_digits=10,
                        null=True,
                        verbose_name="paid amount",
                    ),
                ),
                (
                    "unpaid_amount",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        default=None,
                        editable=False,
                        max_digits=10,
                        null=True,
                        verbose_name="unpaid amount",
                    ),
                ),
                (
                    "overpaid_amount",
                    models.DecimalField(
                        blank=True,
                        db_index=True,
                        decimal_places=2,
                        default=None,
                        editable=False,
                        max_digits=10,
                        null=True,
                        verbose_name="overpaid amount",
                    ),
                ),
                (
                    "close_date",
                    models.DateTimeField(blank=True, db_index=True, default=None, null=True, verbose_name="close date"),
                ),
                (
                    "late_days",
                    models.SmallIntegerField(blank=True, db_index=True, default=None, null=True, verbose_name="late days"),
                ),
                (
                    "state",
                    models.CharField(
                        blank=True,
                        choices=[("N", "Not due yet"), ("D", "Due"), ("L", "Late"), ("P", "Paid")],
                        db_index=True,
                        default="",
                        max_length=1,
                        verbose_name="state",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "invoices",
                "verbose_name": "invoice",
            },
            bases=(models.Model, jutil.cache.CachedFieldsMixin),
        ),
        migrations.AddField(
            model_name="accountentry",
            name="settled_invoice",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="For incoming settlements",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="jacc.Invoice",
                verbose_name="settled invoice",
            ),
        ),
        migrations.AddField(
            model_name="accountentry",
            name="settled_item",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="For matched settlements",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="settlement_set",
                to="jacc.AccountEntry",
                verbose_name="settled item",
            ),
        ),
        migrations.AddField(
            model_name="accountentry",
            name="source",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Source for account entry, e.g. received payment (if any)",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="jacc.AccountEntrySource",
                verbose_name="source",
            ),
        ),
        migrations.AddField(
            model_name="accountentry",
            name="source_invoice",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="For invoice items",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="jacc.Invoice",
                verbose_name="source invoice",
            ),
        ),
        migrations.AddField(
            model_name="accountentry",
            name="type",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="jacc.EntryType",
                verbose_name="type",
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="jacc.AccountType",
                verbose_name="type",
            ),
        ),
    ]
